// ===================================
// Motion Sensor
// ===================================

#include "MPU9250.h"

// an MPU9250 object with the MPU-9250 sensor on I2C bus 0 with address 0x69
MPU9250 IMU(Wire,0x69);

// x/y/z measurements (in the sensor frame, not the lab frame) of acceleration, rotation, and magnetic field
float ax,ay,az,gx,gy,gz,mx,my,mz;
// the horizontal-in-plane component of m, the vertical component of g, the angle between the earings and magnetic north
float mhoriz, gspine, mangle; 
// a measure of the norm of the derivative of the acceleration
float da;

// scaling factors and biases for the magnetic sensor
#define MXSF 1.06
#define MYSF 0.97
#define MZSF 0.97
#define MXB 36.45
#define MYB 12.37
#define MZB -44.42

// Sensor axes
/*      ^       ⊗ az gz
 *    / . \     ⊙ mz
 *   / .   \    ↖ ay gy mx
 *  |   .   |   ↙ ax gx my
 *   \ . . /    ← mhoriz = (mx+my)/√2 
 *    \ . /     ↑ gspine = (gy-gx)/√2 
 *      v       mangle = atan2(mhoriz, mz)
 */

// Low-Pass filter parameters
#define A 0.07
#define B (1-A)

// Data aquisition.
// Put this in the event loop and 100 times a second it will
// read the sensors and update the global variables with sensor
// data.
void imugetdata100Hz() {
  static unsigned long t = 0;
  if (millis()-t<10) return;
  t = millis();
  
  IMU.readSensor();
  
  float tmp;
  tmp = IMU.getAccelX_mss();
  da = B*da + A*(ax-tmp)*(ax-tmp);
  ax = A*tmp+B*ax;
  tmp = IMU.getAccelY_mss();
  da += A*(ay-tmp)*(ay-tmp);
  ay = A*tmp+B*ay;
  tmp = IMU.getAccelZ_mss();
  da += A*(az-tmp)*(az-tmp);
  az = A*tmp+B*az;
  
  gx = A*IMU.getGyroX_rads()+B*gx;
  gy = A*IMU.getGyroY_rads()+B*gy;
  gz = A*IMU.getGyroZ_rads()+B*gz;
  gspine = (gy-gx)/1.414;
  
  mx = A*IMU.getMagX_uT()+B*mx;
  my = A*IMU.getMagY_uT()+B*my;
  mz = A*IMU.getMagZ_uT()+B*mz;
  mhoriz = (mx+my)/1.414;
  mangle = atan2(mhoriz, mz);  
}


// ===================================
// LED Driver - single LED mode
// ===================================

// Which bits on which ports are connected to which common cathode/anode
#define N -1
const int ANODEBIT[] = {4,5,0,1}; //luckily, all are port F
const int CATHODEBITSB[] = {3,7,0,N,N,N,4,N,5,N,N,N};
const int CATHODEBITSC[] = {N,N,N,N,N,N,N,N,N,7,N,N};
const int CATHODEBITSD[] = {N,N,N,6,4,2,N,7,N,N,N,N};
const int CATHODEBITSF[] = {N,N,N,N,N,N,N,N,N,N,7,6};
const int CATHODEPORT[]  = {0,0,0,2,2,2,0,2,0,1,3,3};

// Which LED corresponds to which common cathode/anode
const int LED_2_ANODE[] = {0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3};
const int LED_2_CATHODE[] = {0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3};

// Turn all common cathode/anode pins to OUTPUTs.
void setpins2output() {
  DDRF = 0b11110011;
  DDRD = 0b11010100;
  DDRC = 0b10000000;
  DDRB = 0b10111001;
}

// Write low to the anodes and high to the cathodes to turn the LEDs off.
void clearall() {
  PORTF = 0b11000000;
  PORTD = 0b11010100;
  PORTC = 0b10000000;
  PORTB = 0b10111001;  
}

// Write high to appropriate anode and low to appropriate cathode to turn a single LED on (selected by anode/cathode indices).
void turnsingleledon(int anode, int cathode, int color) {
  int cath_index = cathode*3+color;
  clearall();
  PORTF = 0b11000000 | (1 << ANODEBIT[anode]);
  switch (CATHODEPORT[cath_index]) {
    case 0: PORTB  = ~(1 << CATHODEBITSB[cath_index]); break;
    case 1: PORTC  = ~(1 << CATHODEBITSC[cath_index]); break;
    case 2: PORTD  = ~(1 << CATHODEBITSD[cath_index]); break;
    case 3: PORTF &= ~(1 << CATHODEBITSF[cath_index]); break;
  }
}

// Write high to appropriate anode and low to appropriate cathode to turn a single LED on (selected by LED index).
void turnsingleledon(int led, int color) {
  turnsingleledon(LED_2_ANODE[led], LED_2_CATHODE[led], color);  
}

// Run a simple test animation.
void quicktest() {
  for (int color = 0; color < 3; color++) {
    for (int led = 0; led < 16; led++) {
      turnsingleledon(led,color);
      delay(100);
    }
  }  
  clearall(); 
}


// ===================================
// LED Driver - persistence of vision mode
// ===================================

// The "framebuffer". We use 3bit colors (0 to 7).
uint8_t bgr[16][3];

// Which cathode/anode pair corresponds to which LED.
const int ANODECATHODE_2_LED[] = {0,4,8,12,1,5,9,13,2,6,10,14,3,7,11,15};

// The duration of a single LEDs soft PWM cycle.
#define DELAY delayMicroseconds(25)

// Clear the framebuffer.
void clearframe() {
  for (int led = 0; led < 16; led++)
  for (int color = 0; color < 3; color++)
  bgr[led][color] = 0;
}

// Set the whole framebuffer to a single color.
void uniformframe(int r, int g, int b) {
  for (int led = 0; led < 16; led++) {
    bgr[led][0] = b;
    bgr[led][1] = g;
    bgr[led][2] = r;
  }
}

// Draw a frame once (roughly 30fps).
void drawframe() {
  for (int cycle = 0; cycle < 8; cycle++) {
    for (int anode = 0; anode < 4; anode++) { 
      PORTF = 0b11000000 | (1 << ANODEBIT[anode]);
      for (int cathode = 0; cathode < 4; cathode++) {
        for (int color = 0; color < 3; color++) {
          if (cycle >= bgr[ANODECATHODE_2_LED[anode*4+cathode]][color]) {
            DELAY;
            continue;
          }
          int cath_index = cathode*3+color;
          switch (CATHODEPORT[cath_index]) {
            case 0:
              PORTB  = ~(1 << CATHODEBITSB[cath_index]);
              DELAY;
              PORTB  = 0b10111001; 
              break;
            case 1:
              PORTC  = ~(1 << CATHODEBITSC[cath_index]);
              DELAY;
              PORTC  = 0b10000000;
              break;
            case 2:
              PORTD  = ~(1 << CATHODEBITSD[cath_index]);
              DELAY;
              PORTD  = 0b11010100;
              break;
            case 3:
              PORTF &= ~(1 << CATHODEBITSF[cath_index]);
              DELAY;
              PORTF |= 0b11000000;
              break;
          }     
        }
      }
    PORTF = 0b11000000;
    }
  }
}

// Repeatedly draw frame for fixed duration (blocking/nothing else runs at the same time)
void drawframetimeout(unsigned long timeout) {
  unsigned long t = millis();
  while (millis()-t<timeout) drawframe();
}


// ===================================
// Capacitive button
// ===================================

// A very hackish trick to detect whether there is noise on our unconnected pin.
#define BUTTON_PIN 7
#define BUTTON_WAIT 500
bool buttonpressed() {
  static unsigned long last_measurement = millis();
  static unsigned long last_press = millis();
  static bool last_read = true;
  if (millis()-last_measurement>10) {
    last_measurement = millis();
    bool new_read = digitalRead(BUTTON_PIN);
    bool pressed = last_read xor new_read;
    last_read = new_read;
    pinMode(BUTTON_PIN,INPUT_PULLUP);
    pinMode(BUTTON_PIN,INPUT);
    if (pressed and millis()-last_press>BUTTON_WAIT) {
      last_press = millis();
      return true;
    }
  }
  return false;
}


// ===================================
// Battery reading.
// ===================================

#include <avr/sleep.h>

#define BAT_FULL 4096
#define BAT_EMPTY 3403

// Set up reading the internal voltage reference.
void setADCtoVCC() {
  ADMUX = _BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle  
}

// Read the battery supply voltage.
long readVCC() {
  long result;
  // Read 1.1V reference against AVcc
  ADCSRA |= _BV(ADSC);
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = 1126400L / result; // Back-calculate AVcc in mV
  return result;
}

// If the battery supply voltage is low, shut down the device.
void shutdownifbatlow() {
  static unsigned long t = millis();
  if (millis()-t<100) return;
  t = millis();
  unsigned long v = (readVCC()+readVCC()+readVCC()+readVCC())/4;
  if (v<BAT_EMPTY) {
    clearall();
    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_enable();
    sleep_mode();
  }
}

// ===================================
// Various drawing routines.
// ===================================

// Global variable that specify whether the IMU or framebuffer are used by the current drawing routine.
bool uses_IMU = true;
bool uses_framebuffer = true;

// Color depending on orientation with respect to magnetic North, brightness depending on recent motion.
void motioncompass() {
  uses_IMU = true;
  uses_framebuffer = true;
  float r,b,g;
  if (mangle < -PI/3) {
    r=0;
    b=(mangle+PI)/(PI*2/3);
    g=1-b;
  }
  else if (mangle < PI/3) {
    r=(mangle+PI/3)/(PI*2/3);
    b=1-r;
    g=0;
  }
  else {
    b=0;
    g=(mangle-PI/3)/(PI*2/3);
    r=1-g;
  }
  uint8_t br = constrain(da*r,0,7);
  uint8_t bb = constrain(da*b,0,7);
  uint8_t bg = constrain(da*g,0,7);
  uniformframe(br,bg,bb);
    
  Serial.print(mangle);Serial.print('\t');Serial.println(0);
}

// Purple, brightness depending on recent motion.
void purplemotion() {
  uses_IMU = true;
  uses_framebuffer = true;
  uint8_t b = constrain(da,0,7);
  uniformframe(b,0,b);
  Serial.print(da);Serial.print('\t');Serial.println(0);
}

// White, brightness depending on recent motion.
void whitemotion() {
  uses_IMU = true;
  uses_framebuffer = true;
  uint8_t b = constrain(da,0,7);
  uniformframe(b,b,b);
  Serial.print(da);Serial.print('\t');Serial.println(0);
}

//Cyan, spiral with motion
void cyanspiralsnail() {
  uses_IMU = true;
  uses_framebuffer = true;

  static float snail_location = 2;
  static float motion_measure = 0;

  if (da > 10) {
    motion_measure = 1;
  }
  else {
    motion_measure *= 0.99; //good at ~25frames/sec
  }

  static unsigned long time_last_change = 0;
  
  if (millis()-time_last_change > (20/motion_measure)) {
    clearframe();
    snail_location += .3;
    snail_location = snail_location - floor(snail_location/16)*16;
    time_last_change = millis();
  }

  for (int i=0; i<16; i++) {
    float relative_index = snail_location-i;
    relative_index = relative_index - floor(relative_index/16)*16;
    if (relative_index>8) {relative_index -= 16;}
    bgr[i][0] = constrain(7.1-0.5*relative_index*relative_index, 0, 7);  
    bgr[i][1] = constrain(7.1-0.5*relative_index*relative_index, 0, 7);  
  }

}

// Draw the current battery level.
void drawbatterylevel() {
  uses_IMU = false;
  uses_framebuffer = true;
  long v = (readVCC()+readVCC()+readVCC()+readVCC())/4;
  clearframe();
  for (int i=0; i<9; i++) {
    bgr[15-i][2] = 7;
  }
  for (int i=0; i<7*(v-BAT_EMPTY)/(BAT_FULL-BAT_EMPTY); i++) {
    bgr[15-9-i][0] = i+1;
    bgr[15-9-i][2] = 7-(i+1);
  }
  Serial.println(v);
}

// An array with all drawing routines.
void (*drawingroutines[]) (void) = {cyanspiralsnail, motioncompass, purplemotion, whitemotion, drawbatterylevel};

// The size of the array with drawing routines.
#define NB_DRAWINGROUTINES (sizeof(drawingroutines)/sizeof(drawingroutines[0]))

// Current drawing routine's index.
uint8_t drawingroutine_index = 0;

// ===================================
// Setup and event loop
// ===================================

#define FRAME_RATE 25

void setup() {
  Serial.begin(115200);

  Serial.println(" Set anode/cathode pins.");
  setpins2output();

  Serial.println("Test all LEDs");
  clearall();
  quicktest();
  clearframe();
  
  Serial.print(" Initialize IMU... status ");
  int imu_status;
  for (int i=0; i<5; i++) { // Seems the soldering can be borderline horible.
    imu_status = IMU.begin();
    if (imu_status>0) break;
    Serial.print(" failed attempt ");
    Serial.println(imu_status);
  }
  Serial.println(imu_status);

  uniformframe(7,7,7); // Wait before sending anything to the IMU.
  drawframetimeout(500);

  if (imu_status < 0) {
    for (int i=0; i<3; i++) {
      uniformframe(7,0,0); // Flash red if the IMU is not working.
      drawframetimeout(300);
      uniformframe(0,0,0);
      drawframetimeout(300);
    }
  } else {
    uniformframe(0,0,7); // Flash blue if the IMU is fine.
    drawframetimeout(500);
    Serial.println("Set magnetometer calibration parameters.");
    IMU.setMagCalX(MXB,MXSF);
    IMU.setMagCalY(MYB,MYSF);
    IMU.setMagCalZ(MZB,MZSF);
  }

  Serial.println("Set battery measurement up.");
  setADCtoVCC();
  drawbatterylevel();
  drawframetimeout(1500);
  clearframe();
  clearall();
}

void loop() {
  shutdownifbatlow();
  if (uses_framebuffer) drawframe();
  if (uses_IMU) imugetdata100Hz();

  static unsigned long t = 0;
  if (millis()-t>1000/FRAME_RATE) {
    t = millis();
    drawingroutines[drawingroutine_index]();
  }

  if (buttonpressed()) {
    Serial.print("Button pressed - new routine ");
    drawingroutine_index = (drawingroutine_index+1)%NB_DRAWINGROUTINES;
    Serial.println(drawingroutine_index);
    clearall();
    turnsingleledon(drawingroutine_index,0);
    unsigned long tt = millis(); while (millis()-tt<1000) {buttonpressed();} // We need to continue resetting the button pin during waiting.
    clearall();
    clearframe();
  }
}
